class Stack:
    def __init__(self,N = 'inf'):
        self._data = [] 
        self._N = N  

    # Size of Stack
    def __len__(self):
        return len(self._data)
    # Print stack using print function
    def __str__(self):
        if self.isEmpty():
            return "Stack is Empty!"
        s = ""
        for i in self._data:
            s += str(i) + " "
        return s
    # Check if stack is empty
    def isEmpty(self):
        return len(self._data) == 0
    # Check if stack is full
    def isFull(self):
        return len(self._data) == self._N

    # Push an element to the stack
    def push(self,e):
        if self.isFull():
            print("Stack is Full!")
            return
        else:
            self._data.append(e)

    # Pop an element out of the stack
    def pop(self):
        if self.isEmpty():
            print("Stack is Empty!")
            return
        else:
            return self._data.pop()

    # Return top element of the stack
    def top(self):
        if self.isEmpty():
            print("Stack is Empty!")
            return
        else:
            return self._data[-1]
    
    # Display the Stack
    def display(self):
        for i in self._data:
            print("|",i,end=" ")
        print()
    
    # Delete the stack
    def clear(self):
        self._data = []
    
if __name__ == "__main__":
    s = Stack(5)
    s.push(1)
    s.push(4)
    s.push(2)
    s.push(8)
    s.push(6)
    print(len(s))
    print(s.isEmpty())
    print(s.isFull())
    s.push(5)
    print(s.pop())
    print(s.top())
    s.display()
    print(s)
    s.clear()
    print(s)



class _Node:
    __slots__ = '_element','_next'
    def __init__(self,element,next):
        self._element = element
        self._next = next
    
class Queue:
    def __init__(self,N = 'inf'):
        self._front = None 
        self._rare = None 
        self._N = N
        self._size = 0
    
    # Size of linkedlist
    def __len__(self):
        return self._size
    # Print Queue using print function
    def __str__(self):
        if self.isEmpty():
            return "DEQueue is Empty!"
        s = ""
        p = self._front
        while p:
            s += str(p._element) + " "
            p = p._next
        return s
    # Check if list is empty
    def isEmpty(self):
        return self._size == 0
    # Check if Queue is full
    def isFull(self):
        return self._size == self._N

    # Insert an element to the Queue
    def addFirst(self,e):
        if self.isFull():
            print("DEQueue is Full!")
            return
        newest = _Node(e,None)
        if self.isEmpty():
            self._rare = newest
        else:
            newest._next = self._front
        self._front = newest
        self._size += 1
    def addLast(self,e):
        if self.isFull():
            print("DEQueue is Full!")
            return
        newest = _Node(e,None)
        if self.isEmpty():
            self._front = newest
        else:
            self._rare._next = newest
        self._rare = newest
        self._size += 1

    # Remove an element out of the Queue
    def removeFirst(self):
        if self.isEmpty():
            print("DEQueue is Empty!")
            return
        elif self._size ==1:
            e = self._front._element
            self.clear()
            return e
        else:
            e = self._front._element
            self._front = self._front._next
            self._size -= 1
            return e
    def removeLast(self):
        if self.isEmpty():
            print('DEQueue is empty!')
            return
        elif (self._size==1):
            self.removeFirst()
            return
        p = self._front
        i = 1
        while(i<len(self)-1):
            p = p._next
            i += 1
        self._rare = p
        e = p._next._element
        self._rare._next = None    
        self._size -= 1
        return e
        
    # Return first or last  element of the Queue
    def first(self):
        if self.isEmpty():
            print("DEQueue is Empty!")
            return
        else:
            return self._front._element
    def last(self):
        if self.isEmpty():
            print("DEQueue is Empty!")
            return
        else:
            return self._rare._element

    # Display the Queue
    def display(self):
        if self.isEmpty():
            return "DEQueue is Empty!"
        s = ""
        p = self._front
        while p:
            s += str(p._element)+" | "
            p = p._next
        print(s)
    
    # Delete the Queue
    def clear(self):
        self._size = 0
        self._front = None

if __name__ == "__main__":
    s = Queue(8)
    s.addFirst(1)
    s.addLast(4)
    s.addLast(2)
    s.addFirst(3)
    s.addLast(8)
    s.addFirst(5)
    s.addLast(6)
    s.addFirst(9)
    print(len(s))
    print(s.isEmpty())
    print(s.isFull())
    s.addLast(5)
    s.display()
    print(s.removeFirst())
    s.display()
    print(s.removeLast())
    s.display()
    print(s.first())
    print(s.last())
    print(s)
    s.clear()
    print(s)


def sort(arr,mode=0):
    if mode==0:                                                 # Ascending Order
        l = len(arr)
        for i in range(0,l-1):
            swap = False
            for j in range(0,l-1):  
                if arr[j] > arr[j+1]:
                    arr[j], arr[j+1] = arr[j+1], arr[j] 
                    swap = True
            if swap == False:
                break
        return arr
    elif mode==1:                                               # Descending Order
        l = len(arr)
        for i in range(0,l-1):
            swap = False
            for j in range(0,l-1):  
                if arr[j] < arr[j+1]:
                    arr[j], arr[j+1] = arr[j+1], arr[j] 
                    swap = True
            if swap == False:
                break
        return arr
        
if __name__=='__main__':
    import time
    seconds_1 = time.time()
    a = [9,6,5,8,7,2,3,0,1,4]
    print(sort(a))
    print(sort(a,1))
    seconds = time.time()
    print("runtime:",seconds-seconds_1)


def sort(a,mode=0):
    if mode == 0:                                               # Ascending
        n = len(a)
        mx = max(a)
        digits = len(str(mx))
        l = []
        bins = [l]*10
        for i in range(digits):
            for j in range(n):
                e = int((a[j]/pow(10,i))%10)
                if len(bins[e])>0:
                    bins[e].append(a[j])
                else:
                    bins[e] = [a[j]]
            k = 0
            for x in range(10):
                if len(bins[x])>0:
                    for y in range(len(bins[x])):
                        a[k] = bins[x].pop()
                        k += 1
        return a     
    elif mode == 1:                                             # Descending
        n = len(a)
        mx = max(a)
        digits = len(str(mx))
        l = []
        bins = [l]*10
        for i in range(digits):
            for j in range(n):
                e = int((a[j]/pow(10,i))%10)
                if len(bins[e])>0:
                    bins[e].append(a[j])
                else:
                    bins[e] = [a[j]]
            k = n-1
            for x in range(10):
                if len(bins[x])>0:
                    for y in range(len(bins[x])):
                        a[k] = bins[x].pop()
                        k -= 1
        return a 

if __name__=='__main__':
    import time
    seconds_1 = time.time()
    a = [9,6,5,8,7,2,3,0,1,4]
    print(sort(a))
    print(sort(a,1))
    seconds = time.time()
    print("runtime:",seconds-seconds_1)


def sort(arr,mode=0):
    if mode==0:                                                 # Ascending Order
        l = len(arr)
        gap = l//2
        while gap>0:
            i = gap
            while i<l:
                temp = arr[i]
                j = i-gap
                while j>=0 and arr[j]>temp:
                    arr[j+gap] = arr[j]
                    j = j-gap
                arr[j+gap] = temp
                i += 1
            gap //= 2
        return arr
    elif mode==1:                                               # Descending Order
        l = len(arr)
        gap = l//2
        while gap>0:
            i = gap
            while i<l:
                temp = arr[i]
                j = i-gap
                while j>=0 and arr[j]<temp:
                    arr[j+gap] = arr[j]
                    j = j-gap
                arr[j+gap] = temp
                i += 1
            gap //= 2
        return arr


if __name__=='__main__':
    import time
    seconds_1 = time.time()
    a = [9,6,5,8,7,2,3,0,1,4]
    print(sort(a))
    print(sort(a,1))
    seconds = time.time()
    print("runtime:",seconds-seconds_1)
    


class Queue:
    def __init__(self,N = 'inf'):
        self._data = [] 
        self._N = N  

    # Size of Queue
    def __len__(self):
        return len(self._data)
    # Print Queue using print function
    def __str__(self):
        if self.isEmpty():
            return "Queue is Empty!"
        s = ""
        for i in self._data:
            s += str(i) + " "
        return s
    # Check if Queue is empty
    def isEmpty(self):
        return len(self._data) == 0
    # Check if Queue is full
    def isFull(self):
        return len(self._data) == self._N

    # Insert an element to the Queue
    def enqueue(self,e):
        if self.isFull():
            print("Queue is Full!")
            return
        else:
            self._data.append(e)

    # Remove an element out of the Queue
    def dequeue(self):
        if self.isEmpty():
            print("Queue is Empty!")
            return
        else:
            return self._data.pop(0)

    # Return first element of the Queue
    def first(self):
        if self.isEmpty():
            print("Queue is Empty!")
            return
        else:
            return self._data[0]
    
    # Display the Queue
    def display(self):
        for i in self._data:
            print(i,"|",end=" ")
        print()
    
    # Delete the Queue
    def clear(self):
        self._data = []
    
if __name__ == "__main__":
    s = Queue(5)
    s.enqueue(1)
    s.enqueue(4)
    s.enqueue(2)
    s.enqueue(8)
    s.enqueue(6)
    print(len(s))
    print(s.isEmpty())
    print(s.isFull())
    s.enqueue(5)
    print(s.dequeue())
    print(s.first())
    s.display()
    print(s)
    s.clear()
    print(s)



def sort(arr,mode=0):
    if mode==0:                                 # Ascending Order
        l = len(arr)
        for i in range(1,l):
            c = arr[i]
            p = i 
            while arr[p-1]>c and p>0:
                arr[p] = arr[p-1]  
                p = p-1
            a[p] = c
        return arr
    elif mode==1:                               # Descending Order
        l = len(arr)
        for i in range(1,l):
            c = arr[i]
            p = i 
            while arr[p-1]<c and p>0:
                arr[p] = arr[p-1]  
                p = p-1
            a[p] = c
        return arr
        
if __name__=='__main__':
    import time
    seconds_1 = time.time()
    a = [9,6,5,8,7,2,3,0,1,4]
    print(sort(a))
    print(sort(a,1))
    seconds = time.time()
    print("runtime:",seconds-seconds_1)


class _Node:
    __slots__='_element','_next','_prev'
    def __init__(self,element,next,prev):
        self._element = element
        self._next = next
        self._prev = prev

class LinkedList:
    def __init__(self):
        self._head = None 
        self._tail = None
        self._size = 0

    # Size of linkedlist
    def __len__(self):
        return self._size
    # Print Queue using print function
    def __str__(self):
        self.display()
    # Check if list is empty
    def isEmpty(self):
        return self._size == 0

    # Add a node at end of the linkedlist
    def addLast(self,e):
        newest = _Node(e,None,None)
        if self.isEmpty():
            self._head = newest
        else:
            self._tail._next = newest
            newest._prev = self._tail
        self._tail = newest
        self._size += 1
    # Add a node at starting of linkedlist
    def addFirst(self,e):
        newest = _Node(e,None,None)
        if self.isEmpty():
            self._head = newest
            self._tail = newest
        else:
            newest._next = self._head
            self._head._prev = newest
            self._head = newest
        self._size += 1
    # Add a node at given index of linkedlist
    def addPos(self,e,position):  
        if position>self._size or position<0:
            return None 
        elif (position==0):
            self.addFirst(e)
        elif (position==len(self)):
            self.addLast(e)
        else:
            newest = _Node(e,None,None)
            p = self._head
            i = 0
            while (i<position-1):
                p = p._next
                i += 1
            newest._next = p._next
            newest._prev = p
            p._next._prev = newest
            p._next = newest
            self._size += 1
    
    # Print the linkedlist
    def display(self):
        if self.isEmpty():
            print("List is empty!")
            return
        p = self._head
        while p:
            print(p._element,end=" --> ")
            p = p._next
        print()

    # Print the linkedlist in reverse order
    def displayRev(self):
        if self.isEmpty():
            print("List is empty!")
            return
        p = self._tail
        while (p):
            print(p._element,end=" --> ")
            p = p._prev
        print()

    # Search for a key in the linkedlist. Returns the index
    def search(self,key):
        p = self._head
        index = 0
        while p:
            if p._element == key:
                return index
            p = p._next
            index += 1
        return -1

    # Remove first node from the linkedlist    
    def removeFirst(self):
        if self.isEmpty():
            print('List is Empty!!')
            return
        if self._size==1:
            e = self._head._element
            self._head=None
            self._tail=None
            self._size -= 1
            return e  
        e = self._head._element
        self._head = self._head._next
        self._head._prev = None
        self._size -= 1 
        return e 

    # Remove last node from the linked list
    def removeLast(self):
        if self.isEmpty():
            print("List is Empty!!")
            return
        elif (self._size==1):
            self.removeFirst()
            return
        e = self._tail._element
        self._tail = self._tail._prev
        self._tail._next = None 
        self._size -= 1
        return e 

    # Remove a node from a particular position
    def removePos(self,position):  
        if position == 0:
            self.removeFirst()
        elif position == len(self)-1:
            self.removeLast()
        elif position<0 or position>len(self)-1:
            print("Position out of range")
        else:
            p = self._head
            i = 0
            while (i<position-1):
                p = p._next
                i += 1
            e = p._next._element
            p._next = p._next._next
            p._next._prev = p
            self._size -= 1 
            return e
    # Remove node with a given value
    def removeVal(self,key):
        current = self._head
        prev = self._head
        if key == self._head._element:
            self.removeFirst()
            return
        elif key== self._tail._element:
            self.removeLast()
            return
        while current._next != None :
            if current._element == key:
                prev._next = current._next
                current._next._prev = prev
                self._size -= 1 
                return 
            else:
                prev = current
                current = current._next 
        print('Value not in list!')

    # Delete the entire linkedlist
    def clear(self):
        self._head = None 
        self._size = 0
        return "List is cleared!"

if __name__ == '__main__':
    L = LinkedList()
    L2 = LinkedList()
    L.addLast(7)
    L.addLast(4)
    L.addLast(8)
    L.addLast(1)
    L.addFirst(2)
    L.addLast(5)
    L.addLast(12)
    L.addPos(9,0)
    L.display()
    L.displayRev()

    L.removeVal(9)
    L.removeVal(12)
    L.removeVal(8)
    L.removeVal(14)
    L.display()
    

    
    x = L.removeFirst()
    print("Element Removed:",x)
    L.display()

    x = L.removeLast()
    print("Element Removed:",x)
    L.display()

    x = L.removePos(4)
    print("Element Removed:",x)
    L.display()
    L2.removeFirst()
    print(L.search(5))
    print(L.search(8))
    print("Size:",len(L))
    print(L.clear())
    L.display()


from setuptools import setup, find_packages

classifiers=[
'Development Status :: 5 - Production/Stable', 
'Intended Audience :: Education',
'Operating System :: OS Independent',
'License :: OSI Approved :: MIT License',
'Programming Language :: Python :: 3'
]

setup(
    name= 'DSAinPython', 
    version='0.0.5',
    author='Armaan Chauhan',
    author_email='armaanchauhan268@gmail.com',
    description='This Package contains Data Structures and Algorithms', 
    long_description=open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Armaan-268/DSA_in_Python',
    license="MIT",
    classifier=classifiers, 
    keywords='datastructures,algorithms',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['']
)